Utils <- class {}

/// getNthItem
///
/// @param list - The list to get the item from.
/// @param n - The order of the item.
/// @return - The n-th item of the list, null if not found.
function Utils::getNthItem(list, n) {
    if(list.Count() == 0) {
        AILog.Warning("getNthItem: list is empty!");
        return null;
    }

    if(n >= list.Count()) {
        AILog.Warning("getNthItem: list is too short!");
        return null;
    }

    for (local item = list.Begin(), i = 0; list.HasNext(); item = list.Next(), ++i) {
        if(i == n) {
            //AILog.Info("getNthItem: Found: " + item + " " + list.GetValue(item));
            return item;
        }
    }

    return null;
}

function Utils::getValidOffsetTile(tile, offsetX, offsetY) {
    local oldX = AIMap.GetTileX(tile);
	local oldY = AIMap.GetTileY(tile);
	
	local newX = oldX;
	local newY = oldY;
	
	if (offsetX > 0) {
	    for (local x = offsetX; x > 0; x--) {
		    if (AIMap.IsValidTile(AIMap.GetTileIndex(newX + 1, newY))) {
			    newX = newX + 1;
			}
		}
	} else {
	    for (local x = offsetX; x < 0; x++) {
		    if (AIMap.IsValidTile(AIMap.GetTileIndex(newX - 1, newY))) {
			    newX = newX - 1;
			}
		}
	}
	
	if (offsetY > 0) {
	    for (local y = offsetY; y > 0; y--) {
		    if (AIMap.IsValidTile(AIMap.GetTileIndex(newX, newY + 1))) {
			    newY = newY + 1;
			}
		}
	} else {
	    for (local y = offsetY; y < 0; y++) {
		    if (AIMap.IsValidTile(AIMap.GetTileIndex(newX, newY - 1))) {
			    newY = newY - 1;
			}
		}
	}
	
	return AIMap.GetTileIndex(newX, newY);
}

/// getOffsetTile
///
/// @param tile - The starting tile.
/// @param offsetX - The x-axis offset.
/// @param offsetY - The y-axis offset.
/// @return - The offset tile.
function Utils::getOffsetTile(tile, offsetX, offsetY) {
    local oldX = AIMap.GetTileX(tile);
    local oldY = AIMap.GetTileY(tile);

    local newX = oldX + offsetX;
    local newY = oldY + offsetY;

    if(((newX < 1) || (newY < 1)) // 0 if freeform_edges off
        && (((newX > AIMap.GetMapSizeX() - 2) || (newY > AIMap.GetMapSizeY() - 2)))) {
        return AIMap.TILE_INVALID;
    }

    return AIMap.GetTileIndex(newX, newY);
}

/// getAdjacentTiles
///
/// @param tile - The starting tile.
/// @return - The AITileList of adjacent tiles.
function Utils::getAdjacentTiles(tile) {
    local adjTiles = AITileList();

    local offsetTile = Utils.getOffsetTile(tile, 0, -1)
    if(offsetTile != AIMap.TILE_INVALID) {
        adjTiles.AddTile(offsetTile);
    }

    offsetTile = Utils.getOffsetTile(tile, 1, 0)
    if(offsetTile != AIMap.TILE_INVALID) {
        adjTiles.AddTile(offsetTile);
    }

    offsetTile = Utils.getOffsetTile(tile, 0, 1)
    if(offsetTile != AIMap.TILE_INVALID) {
        adjTiles.AddTile(offsetTile);
    }

    offsetTile = Utils.getOffsetTile(tile, -1, 0)
    if(offsetTile != AIMap.TILE_INVALID) {
        adjTiles.AddTile(offsetTile);
    }

    return adjTiles;
};

function Utils::IsStationBuildableTile(tile) {
    if(AITile.GetSlope(tile) == AITile.SLOPE_FLAT && (AITile.IsBuildable(tile) || AIRoad.IsRoadTile(tile))) {
        return true;
	}
	return false;
}

function Utils::AreOtherStationsNearby(tile, cargoClass, stationId) {
    local stationType = cargoClass == AICargo.CC_PASSENGERS ? AIStation.STATION_BUS_STOP : AIStation.STATION_TRUCK_STOP;
	
    //check if there are other stations squareSize squares nearby
    local squareSize = AIStation.GetCoverageRadius(stationType);
	if (stationId == null) {
	    squareSize = squareSize * 2;
	}

	local square = AITileList();
    if(!AIController.GetSetting("is_friendly")) {
        //dont care about enemy stations when is_friendly is off
        square.AddRectangle(Utils.getValidOffsetTile(tile, (-1) * squareSize, (-1) * squareSize),
            Utils.getValidOffsetTile(tile, squareSize, squareSize));

        //if another road station of mine is nearby return true
        for(local tile = square.Begin(); square.HasNext(); tile = square.Next()) {
            if(Utils.isTileMyRoadStation(tile, cargoClass)) { //negate second expression to merge your stations
                return true;
        	}
        }
    }
	else {
        square.AddRectangle(Utils.getValidOffsetTile(tile, (-1) * squareSize, (-1) * squareSize),
            Utils.getValidOffsetTile(tile, squareSize, squareSize));

        //if any other station is nearby, except my own airports, return true
        for (local tile = square.Begin(); square.HasNext(); tile = square.Next()) {
            if(AITile.IsStationTile(tile)) {
			    if (AITile.GetOwner(tile) != AICompany.ResolveCompanyID(AICompany.COMPANY_SELF)) {
			        return true;
				} else {
			        local stationTiles = AITileList_StationType(AIStation.GetStationID(tile), stationType);
				    if (stationTiles.HasItem(tile)) {
                        return true;
					}
				}
            }
        }
    }
	
	return false;
};

/// valuateTruckStopTile
///
/// @param tile - The starting tile to valuate.
/// @param cargoClass - should be either AICargo.CC_MAIL, or AICargo.CC_PASSENGERS to get correct cargo acceptance.
/// @return - The tile value.
function Utils::valuateTruckStopTile(tile, cargoClass, stationId, cargoType) {
    local radiusType;
    if(cargoClass == AICargo.CC_MAIL) {
        radiusType = AIStation.STATION_TRUCK_STOP;
    }
    else {
        radiusType = AIStation.STATION_BUS_STOP;
    }

    if(AITile.GetCargoAcceptance(tile,
        cargoType,
        1,
        1,
        AIStation.GetCoverageRadius(radiusType)) < 8) {
            return 0;
    }

    local value = AITile.GetCargoProduction(tile,
                cargoType,
                1,
                1,
                AIStation.GetCoverageRadius(radiusType));
    return value;
}

function Utils::isTileMyRoadStation(tile, cargoClass) {
    if(AITile.IsStationTile(tile) && AITile.GetOwner(tile) == AICompany.ResolveCompanyID(AICompany.COMPANY_SELF) &&
	    AIStation.HasStationType(AIStation.GetStationID(tile), cargoClass == AICargo.CC_PASSENGERS ? AIStation.STATION_BUS_STOP : AIStation.STATION_TRUCK_STOP)) {
        return 1;
    }

    return 0;
}

function Utils::isTileMyAirport(tile) {
    if (AITile.IsStationTile(tile) && AITile.GetOwner(tile) == AICompany.ResolveCompanyID(AICompany.COMPANY_SELF) &&
	    AIStation.HasStationType(AIStation.GetStationID(tile), AIStation.STATION_AIRPORT)) {
	    return 1;
	}
	
	return 0;
}

function Utils::isTileMyStationWithoutAirport(tile) { // Checks RoadStation
    if(AITile.IsStationTile(tile) && AITile.GetOwner(tile) == AICompany.ResolveCompanyID(AICompany.COMPANY_SELF) &&
	    !AIStation.HasStationType(AIStation.GetStationID(tile), AIStation.STATION_AIRPORT)) {
        return 1;
    }

    return 0;
}

function Utils::isTileMyStationWithoutRoadStation(tile, cargoClass) { // Checks Airport
    if (AITile.IsStationTile(tile) && AITile.GetOwner(tile) == AICompany.ResolveCompanyID(AICompany.COMPANY_SELF) &&
	    !AIStation.HasStationType(AIStation.GetStationID(tile), cargoClass == AICargo.CC_PASSENGERS ? AIStation.STATION_BUS_STOP : AIStation.STATION_TRUCK_STOP)) {
		return 1;
	}
	
	return 0;
}

/// valuateTownRoadEnds - Creates a square around @tile and finds all tiles with one adjacent road tile.
///
/// @param tile - The starting tile to valuate.
/// @return - True if road ends on the tile, false otherwise.
function Utils::valuateTownRoadEnds(tile) {
    if(!AIRoad.IsRoadTile(tile)) {
        return 0;
    }

    local temp = AITileList();
    temp.AddList(getAdjacentTiles(tile));
    temp.Valuate(AIRoad.IsRoadTile);
    temp.KeepValue(1);

    if(temp.Count() == 1) {
        return 1;
    }

    return 0;
}


/// getCargoId - Returns either mail cargo id, or passenger cargo id.
///
/// @param cargoClass - either AICargo.CC_MAIL, or AICargo.CC_PASSENGERS
/// @return - Cargo list.
function Utils::getCargoId(cargoClass) {
    local cargoList = AICargoList();
    cargoList.Valuate(AICargo.HasCargoClass, cargoClass);
    cargoList.KeepValue(1);

    //both AICargo.CC_MAIL and AICargo.CC_PASSENGERS should have only one cargo
    return cargoList.Begin();
}

function Utils::checkAdjacentAirport(stationTile, cargoClass, stationId)
{
    if (stationId != AIStation.STATION_NEW) {
	    return stationId;
	}
	
	if (!AIController.GetSetting("station_spread")) {
	    return AIStation.STATION_NEW;
	}
	
	local spread_rad = AIGameSettings.GetValue("station_spread") - 1;

    local tileList = AITileList();
	local spreadrectangle = [Utils.getValidOffsetTile(stationTile, (-1) * spread_rad, (-1) * spread_rad), Utils.getValidOffsetTile(stationTile, spread_rad, spread_rad)];
	tileList.AddRectangle(spreadrectangle[0], spreadrectangle[1]);
	
	tileList.Valuate(Utils.isTileMyStationWithoutRoadStation, cargoClass);
	tileList.KeepValue(1);
	tileList.Valuate(AIStation.GetStationID);
	
	local airportList = AIList();
	
	for (local tile = tileList.Begin(); tileList.HasNext(); tileList.Next()) {
		airportList.AddItem(tileList.GetValue(tile), AITile.GetDistanceManhattanToTile(tile, stationTile));
	}
	
	local spreadrectangle_top_x = AIMap.GetTileX(spreadrectangle[0]);
	local spreadrectangle_top_y = AIMap.GetTileY(spreadrectangle[0]);
	local spreadrectangle_bot_x = AIMap.GetTileX(spreadrectangle[1]);
	local spreadrectangle_bot_y = AIMap.GetTileY(spreadrectangle[1]);
	
	local list = AIList();
	list.AddList(airportList);
	for (local airportId = airportList.Begin(); airportList.HasNext(); airportId = airportList.Next()) {
	    local airportTiles = AITileList_StationType(airportId, AIStation.STATION_ANY);
	    local airport_top_x = AIMap.GetTileX(AIBaseStation.GetLocation(airportId));
		local airport_top_y = AIMap.GetTileY(AIBaseStation.GetLocation(airportId));
		local airport_bot_x = airport_top_x;
		local airport_bot_y = airport_top_y;
		for (local tile = airportTiles.Begin(); airportTiles.HasNext(); tile = airportTiles.Next()) {
		    local tile_x = AIMap.GetTileX(tile);
			local tile_y = AIMap.GetTileY(tile);
		    if (tile_x < airport_top_x) {
			    airport_top_x = tile_x;
			}
			if (tile_x > airport_bot_x) {
			    airport_bot_x = tile_x;
			}
			if (tile_y < airport_top_y) {
			    airport_top_y = tile_y;
			}
			if (tile_y > airport_bot_y) {
			    airport_bot_y = tile_y;
			}
		}
		
		if (spreadrectangle_top_x > airport_top_x || 
		    spreadrectangle_top_y > airport_top_y ||
			spreadrectangle_bot_x < airport_bot_x ||
			spreadrectangle_bot_y < airport_bot_y) {
			list.RemoveItem(airportId);
		}
	}
	list.Sort(AIList.SORT_BY_VALUE, true);
	
	local adjacentStation = AIStation.STATION_NEW;
	if(list.Count()) {
		adjacentStation = list.Begin();
//		AILog.Info("adjacentStation = " + AIStation.GetName(adjacentStation) + " ; stationtTile = " + AIMap.GetTileX(stationTile) + "," + AIMap.GetTileY(stationTile));
	}

	return adjacentStation;
}

/**
 * Check if we have enough money (via loan and on bank).
 */
function Utils::HasMoney(money)
{
	local loan_amount = AICompany.GetLoanAmount();
	local max_loan_amount = AICompany.GetMaxLoanAmount();
	local bank_balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	if (bank_balance + max_loan_amount - loan_amount >= money) return true;
	return false;
}

function Utils::GetMoney(money)
{
	local bank_balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	if (bank_balance >= money) return;
	local request_loan = money - bank_balance;
	local loan_interval = AICompany.GetLoanInterval();
	local over_interval = request_loan % loan_interval;
	request_loan += loan_interval - over_interval;
	local loan_amount = AICompany.GetLoanAmount();
	local max_loan_amount = AICompany.GetMaxLoanAmount();
	if (loan_amount + request_loan > max_loan_amount) {
	    AICompany.SetLoanAmount(max_loan_amount);
	} else {
	    AICompany.SetLoanAmount(loan_amount + request_loan);
	}
}

function Utils::RepayLoan()
{
    local bank_balance = AICompany.GetBankBalance(AICompany.COMPANY_SELF);
	local loan_amount = AICompany.GetLoanAmount();
	local repay_loan = loan_amount - bank_balance > 0 ? loan_amount - bank_balance : 0;
	AICompany.SetMinimumLoanAmount(repay_loan);
}

class MoneyTest {
    function DoMoneyTest() {
	    local price = GetPrice();
        if (Utils.HasMoney(price)) {
		    Utils.GetMoney(price);
        }
        if (DoAction()) {
            Utils.RepayLoan();
            return true;
        }
//		AILog.Error(AIError.GetLastErrorString());
        Utils.RepayLoan();
        return false;
    }
}